# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
'''
Provide grib 1 and 2 phenomenon translations to + from CF terms.

This is done by wrapping '_grib_cf_map.py',
which is in a format provided by the metadata translation project.

Currently supports only these ones:

* grib1 --> cf
* grib2 --> cf
* cf --> grib2

'''
from collections import namedtuple
import re
import warnings

import cf_units

from . import _grib_cf_map as grcf
import iris.std_names


class _LookupTable(dict):
    """
    Specialised dictionary object for making lookup tables.

    Returns None for unknown keys (instead of raising exception).
    Raises exception for any attempt to change an existing entry,
    (but it is still possible to remove keys)

    """
    def __init__(self, *args, **kwargs):
        self._super = super()
        self._super.__init__(*args, **kwargs)

    def __getitem__(self, key):
        if key not in self:
            return None
        return self._super.__getitem__(key)

    def __setitem__(self, key, value):
        if key in self and self[key] is not value:
            raise KeyError('Attempted to set dict[{}] = {}, '
                           'but this is already set to {}.'.format(
                               key, value, self[key]))
        self._super.__setitem__(key, value)


# Define namedtuples for keys+values of the Grib1 lookup table.

_Grib1ToCfKeyClass = namedtuple(
    'Grib1CfKey',
    ('table2_version', 'centre_number', 'param_number'))

# NOTE: this form is currently used for both Grib1 *and* Grib2
_GribToCfDataClass = namedtuple(
    'Grib1CfData',
    ('standard_name', 'long_name', 'units', 'set_height'))


# Create the grib1-to-cf lookup table.

def _make_grib1_cf_table():
    """ Build the Grib1 to CF phenomenon translation table. """
    table = _LookupTable()

    def _make_grib1_cf_entry(table2_version, centre_number, param_number,
                             standard_name, long_name, units, set_height=None):
        """
        Check data, convert types and create a new _GRIB1_CF_TABLE key/value.

        Note that set_height is an optional parameter.  Used to denote
        phenomena that imply a height definition (agl),
        e.g. "2-metre tempererature".

        """
        grib1_key = _Grib1ToCfKeyClass(table2_version=int(table2_version),
                                       centre_number=int(centre_number),
                                       param_number=int(param_number))
        if standard_name is not None:
            if standard_name not in iris.std_names.STD_NAMES:
                warnings.warn('{} is not a recognised CF standard name '
                              '(skipping).'.format(standard_name))
                return None
        # convert units string to iris Unit (i.e. mainly, check it is good)
        a_cf_unit = cf_units.Unit(units)
        cf_data = _GribToCfDataClass(standard_name=standard_name,
                                     long_name=long_name,
                                     units=a_cf_unit,
                                     set_height=set_height)
        return (grib1_key, cf_data)

    # Interpret the imported Grib1-to-CF table.
    for (grib1data, cfdata) in grcf.GRIB1_LOCAL_TO_CF.items():
        assert grib1data.edition == 1
        association_entry = _make_grib1_cf_entry(
            table2_version=grib1data.t2version,
            centre_number=grib1data.centre,
            param_number=grib1data.iParam,
            standard_name=cfdata.standard_name,
            long_name=cfdata.long_name,
            units=cfdata.units)
        if association_entry is not None:
            key, value = association_entry
            table[key] = value

    # Do the same for special Grib1 codes that include an implied height level.
    for (grib1data, (cfdata, extra_dimcoord)) \
            in grcf.GRIB1_LOCAL_TO_CF_CONSTRAINED.items():
        assert grib1data.edition == 1
        if extra_dimcoord.standard_name != 'height':
            raise ValueError('Got implied dimension coord of "{}", '
                             'currently can only handle "height".'.format(
                                 extra_dimcoord.standard_name))
        if extra_dimcoord.units != 'm':
            raise ValueError('Got implied dimension units of "{}", '
                             'currently can only handle "m".'.format(
                                 extra_dimcoord.units))
        if len(extra_dimcoord.points) != 1:
            raise ValueError('Implied dimension has {} points, '
                             'currently can only handle 1.'.format(
                                 len(extra_dimcoord.points)))
        association_entry = _make_grib1_cf_entry(
            table2_version=int(grib1data.t2version),
            centre_number=int(grib1data.centre),
            param_number=int(grib1data.iParam),
            standard_name=cfdata.standard_name,
            long_name=cfdata.long_name,
            units=cfdata.units,
            set_height=extra_dimcoord.points[0])
        if association_entry is not None:
            key, value = association_entry
            table[key] = value

    return table


_GRIB1_CF_TABLE = _make_grib1_cf_table()


# Define a namedtuple for the keys of the Grib2 lookup table.

_Grib2ToCfKeyClass = namedtuple(
    'Grib2CfKey',
    ('param_discipline', 'param_category', 'param_number'))


# Create the grib2-to-cf lookup table.

def _make_grib2_to_cf_table():
    """ Build the Grib2 to CF phenomenon translation table. """
    table = _LookupTable()

    def _make_grib2_cf_entry(param_discipline, param_category, param_number,
                             standard_name, long_name, units):
        """
        Check data, convert types and make a _GRIB2_CF_TABLE key/value pair.

        Note that set_height is an optional parameter.  Used to denote
        phenomena that imply a height definition (agl),
        e.g. "2-metre tempererature".

        """
        grib2_key = _Grib2ToCfKeyClass(param_discipline=int(param_discipline),
                                       param_category=int(param_category),
                                       param_number=int(param_number))
        if standard_name is not None:
            if standard_name not in iris.std_names.STD_NAMES:
                warnings.warn('{} is not a recognised CF standard name '
                              '(skipping).'.format(standard_name))
                return None
        # convert units string to iris Unit (i.e. mainly, check it is good)
        a_cf_unit = cf_units.Unit(units)
        cf_data = _GribToCfDataClass(standard_name=standard_name,
                                     long_name=long_name,
                                     units=a_cf_unit,
                                     set_height=None)
        return (grib2_key, cf_data)

    # Interpret the grib2 info from grib_cf_map
    for grib2data, cfdata in grcf.GRIB2_TO_CF.items():
        assert grib2data.edition == 2
        association_entry = _make_grib2_cf_entry(
            param_discipline=grib2data.discipline,
            param_category=grib2data.category,
            param_number=grib2data.number,
            standard_name=cfdata.standard_name,
            long_name=cfdata.long_name,
            units=cfdata.units)
        if association_entry is not None:
            key, value = association_entry
            table[key] = value

    return table


_GRIB2_CF_TABLE = _make_grib2_to_cf_table()


# Define namedtuples for key+values of the cf-to-grib2 lookup table.

_CfToGrib2KeyClass = namedtuple(
    'CfGrib2Key',
    ('standard_name', 'long_name'))

_CfToGrib2DataClass = namedtuple(
    'CfGrib2Data',
    ('discipline', 'category', 'number', 'units'))


# Create the cf-to-grib2 lookup table.

def _make_cf_to_grib2_table():
    """ Build the Grib1 to CF phenomenon translation table. """
    table = _LookupTable()

    def _make_cf_grib2_entry(standard_name, long_name,
                             param_discipline, param_category, param_number,
                             units):
        """
        Check data, convert types and make a new _CF_TABLE key/value pair.

        """
        assert standard_name is not None or long_name is not None
        if standard_name is not None:
            long_name = None
            if standard_name not in iris.std_names.STD_NAMES:
                warnings.warn('{} is not a recognised CF standard name '
                              '(skipping).'.format(standard_name))
                return None
        cf_key = _CfToGrib2KeyClass(standard_name, long_name)
        # convert units string to iris Unit (i.e. mainly, check it is good)
        a_cf_unit = cf_units.Unit(units)
        grib2_data = _CfToGrib2DataClass(discipline=int(param_discipline),
                                         category=int(param_category),
                                         number=int(param_number),
                                         units=a_cf_unit)
        return (cf_key, grib2_data)

    # Interpret the imported CF-to-Grib2 table into a lookup table
    for cfdata, grib2data in grcf.CF_TO_GRIB2.items():
        assert grib2data.edition == 2
        a_cf_unit = cf_units.Unit(cfdata.units)
        association_entry = _make_cf_grib2_entry(
            standard_name=cfdata.standard_name,
            long_name=cfdata.long_name,
            param_discipline=grib2data.discipline,
            param_category=grib2data.category,
            param_number=grib2data.number,
            units=a_cf_unit)
        if association_entry is not None:
            key, value = association_entry
            table[key] = value

    return table

_CF_GRIB2_TABLE = _make_cf_to_grib2_table()


# Interface functions for translation lookup

def grib1_phenom_to_cf_info(table2_version, centre_number, param_number):
    """
    Lookup grib-1 parameter --> cf_data or None.

    Returned cf_data has attributes:

    * standard_name
    * long_name
    * units : a :class:`cf_units.Unit`
    * set_height :  a scalar 'height' value , or None

    """
    grib1_key = _Grib1ToCfKeyClass(table2_version=table2_version,
                                   centre_number=centre_number,
                                   param_number=param_number)
    return _GRIB1_CF_TABLE[grib1_key]


def grib2_phenom_to_cf_info(param_discipline, param_category, param_number):
    """
    Lookup grib-2 parameter --> cf_data or None.

    Returned cf_data has attributes:

    * standard_name
    * long_name
    * units : a :class:`cf_units.Unit`

    """
    grib2_key = _Grib2ToCfKeyClass(param_discipline=int(param_discipline),
                                   param_category=int(param_category),
                                   param_number=int(param_number))
    return _GRIB2_CF_TABLE[grib2_key]


def cf_phenom_to_grib2_info(standard_name, long_name=None):
    """
    Lookup CF names --> grib2_data or None.

    Returned grib2_data has attributes:

    * discipline
    * category
    * number
    * units : a :class:`cf_units.Unit`
        The unit represents the defined reference units for the message data.

    """
    if standard_name is not None:
        long_name = None
    return _CF_GRIB2_TABLE[(standard_name, long_name)]


class GRIBCode(namedtuple('GRIBCode',
                          'edition discipline category number')):
    """
    An object representing a specific Grib phenomenon identity.

    Basically a namedtuple of (edition, discipline, category, number).

    Also provides a string representation, and supports creation from: another
    similar object; a tuple of numbers; or any string with 4 separate decimal
    numbers in it.

    """
    __slots__ = ()

    def __new__(cls, edition_or_string,
                discipline=None, category=None, number=None):
        args = (edition_or_string, discipline, category, number)
        nargs = sum(arg is not None for arg in args)
        if nargs == 1:
            # Single argument: convert to a string and extract 4 integers.
            # NOTE: this also allows input from a GRIBCode, or a plain tuple.
            edition_or_string = str(edition_or_string)
            edition, discipline, category, number = \
                cls._fournums_from_gribcode_string(edition_or_string)
        elif nargs == 4:
            edition = edition_or_string
            edition, discipline, category, number = [
                int(arg)
                for arg in (edition, discipline, category, number)]
        else:
            msg = ('Cannot create GRIBCode from {} arguments, '
                   '"GRIBCode{!r}" : '
                   'expected either 1 or 4 non-None arguments.')
            raise ValueError(msg.format(nargs, args))

        return super(GRIBCode, cls).__new__(
            cls, edition, discipline, category, number)

    RE_PARSE_FOURNUMS = re.compile(4 * r'[^\d]*(\d*)')

    @classmethod
    def _fournums_from_gribcode_string(cls, edcn_string):
        parsed_ok = False
        nums_match = cls.RE_PARSE_FOURNUMS.match(edcn_string).groups()
        if nums_match is not None:
            try:
                nums = [int(grp) for grp in nums_match]
                parsed_ok = True
            except ValueError:
                pass

        if not parsed_ok:
            msg = ('Invalid argument for GRIBCode creation, '
                   '"GRIBCode({!r})" : '
                   'requires 4 numbers, separated by non-numerals.')
            raise ValueError(msg.format(edcn_string))

        return nums

    PRINT_FORMAT = 'GRIB{:1d}:d{:03d}c{:03d}n{:03d}'

    def __str__(self):
        result = self.PRINT_FORMAT.format(
            self.edition, self.discipline, self.category, self.number)
        return result



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test that iris load callbacks work with iris-grib.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests


import iris
import iris.coords


class TestCallbacks(tests.IrisGribTest):
    def test_load_callback(self):
        # What this actually tests:
        #    1. iris.load works with grib (though the GRIB picker is in Iris)
        #    2. callbacks work with the grib loader
        #    3. grib loaded result matches a cube snapshot
        def load_callback(cube, field, filename):
            # GRIB2 loader callback : 'field' is a GribMessage, which
            # has 'sections'.
            cube.add_aux_coord(
                iris.coords.AuxCoord(
                    field.sections[1]["year"],
                    long_name="extra_year_number_coord",
                    units="no_unit",
                )
            )

        fname = tests.get_data_path(("GRIB", "global_t", "global.grib2"))
        cube = iris.load_cube(fname, callback=load_callback)
        self.assertCML(cube)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :func:`iris_grib._grib1_load_rules.grib1_convert`."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else
import iris_grib.tests as tests

import gribapi
from unittest import mock

from iris.aux_factory import HybridPressureFactory
from iris.exceptions import TranslationError
from iris.fileformats.rules import Reference

from iris_grib import GribWrapper
from iris_grib._grib1_load_rules import grib1_convert
from iris_grib.tests.unit import TestField


class TestBadEdition(tests.IrisGribTest):
    def test(self):
        message = mock.Mock(edition=2)
        emsg = 'GRIB edition 2 is not supported'
        with self.assertRaisesRegex(TranslationError, emsg):
            grib1_convert(message)


class TestBoundedTime(TestField):
    @staticmethod
    def is_forecast_period(coord):
        return (coord.standard_name == 'forecast_period' and
                coord.units == 'hours')

    @staticmethod
    def is_time(coord):
        return (coord.standard_name == 'time' and
                coord.units == 'hours since epoch')

    def assert_bounded_message(self, **kwargs):
        attributes = {'productDefinitionTemplateNumber': 0,
                      'edition': 1, '_forecastTime': 15,
                      '_forecastTimeUnit': 'hours',
                      'phenomenon_bounds': lambda u: (80, 120),
                      '_phenomenonDateTime': -1,
                      'table2Version': 9999,
                      '_originatingCentre': 'xxx',
                      }
        attributes.update(kwargs)
        message = mock.Mock(**attributes)
        self._test_for_coord(message, grib1_convert, self.is_forecast_period,
                             expected_points=[35],
                             expected_bounds=[[15, 55]])
        self._test_for_coord(message, grib1_convert, self.is_time,
                             expected_points=[100],
                             expected_bounds=[[80, 120]])

    def test_time_range_indicator_2(self):
        self.assert_bounded_message(timeRangeIndicator=2)

    def test_time_range_indicator_3(self):
        self.assert_bounded_message(timeRangeIndicator=3)

    def test_time_range_indicator_4(self):
        self.assert_bounded_message(timeRangeIndicator=4)

    def test_time_range_indicator_5(self):
        self.assert_bounded_message(timeRangeIndicator=5)

    def test_time_range_indicator_51(self):
        self.assert_bounded_message(timeRangeIndicator=51)

    def test_time_range_indicator_113(self):
        self.assert_bounded_message(timeRangeIndicator=113)

    def test_time_range_indicator_114(self):
        self.assert_bounded_message(timeRangeIndicator=114)

    def test_time_range_indicator_115(self):
        self.assert_bounded_message(timeRangeIndicator=115)

    def test_time_range_indicator_116(self):
        self.assert_bounded_message(timeRangeIndicator=116)

    def test_time_range_indicator_117(self):
        self.assert_bounded_message(timeRangeIndicator=117)

    def test_time_range_indicator_118(self):
        self.assert_bounded_message(timeRangeIndicator=118)

    def test_time_range_indicator_123(self):
        self.assert_bounded_message(timeRangeIndicator=123)

    def test_time_range_indicator_124(self):
        self.assert_bounded_message(timeRangeIndicator=124)

    def test_time_range_indicator_125(self):
        self.assert_bounded_message(timeRangeIndicator=125)


class Test_GribLevels(tests.IrisTest):
    def test_grib1_hybrid_height(self):
        gm = gribapi.grib_new_from_samples('regular_gg_ml_grib1')
        gw = GribWrapper(gm)
        results = grib1_convert(gw)

        factory, = results[0]
        self.assertEqual(factory.factory_class, HybridPressureFactory)
        delta, sigma, ref = factory.args
        self.assertEqual(delta, {'long_name': 'level_pressure'})
        self.assertEqual(sigma, {'long_name': 'sigma'})
        self.assertEqual(ref, Reference(name='surface_pressure'))

        coords_and_dims = results[8]
        coord, = [co for co, _ in coords_and_dims
                  if co.name() == 'model_level_number']
        self.assertEqual(coord.units, '1')
        self.assertEqual(coord.attributes['positive'], 'up')
        coord, = [co for co, _ in coords_and_dims
                  if co.name() == 'level_pressure']
        self.assertEqual(coord.units, 'Pa')
        coord, = [co for co, _ in coords_and_dims
                  if co.name() == 'sigma']
        self.assertEqual(coord.units, '1')


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration tests for format interoperability.

Checks that specific features of cubes loaded from other formats are correctly
represented by grib saving.

"""



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration test for round-trip loading and saving of various grids.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris import load_cube, save
from iris.coord_systems import RotatedGeogCS
from iris.fileformats.pp import EARTH_RADIUS as UM_DEFAULT_EARTH_RADIUS
from iris.util import is_regular

from iris_grib.grib_phenom_translation import GRIBCode


class TestGDT5(tests.TestGribMessage):
    def test_save_load(self):
        # Load sample UKV data (variable-resolution rotated grid).
        path = tests.get_data_path(("PP", "ukV1", "ukVpmslont.pp"))
        cube = load_cube(path)

        # Extract a single 2D field, for simplicity.
        self.assertEqual(cube.ndim, 3)
        self.assertEqual(cube.coord_dims("time"), (0,))
        cube = cube[0]

        # Check that it has a rotated-pole variable-spaced grid, as expected.
        x_coord = cube.coord(axis="x")
        self.assertIsInstance(x_coord.coord_system, RotatedGeogCS)
        self.assertFalse(is_regular(x_coord))

        # Write to temporary file, check that key contents are in the file,
        # then load back in.
        with self.temp_filename("ukv_sample.grib2") as temp_file_path:
            save(cube, temp_file_path)

            # Check that various aspects of the saved file are as expected.
            expect_values = (
                (0, "editionNumber", 2),
                (3, "gridDefinitionTemplateNumber", 5),
                (3, "Ni", cube.shape[-1]),
                (3, "Nj", cube.shape[-2]),
                (3, "shapeOfTheEarth", 1),
                (
                    3,
                    "scaledValueOfRadiusOfSphericalEarth",
                    int(UM_DEFAULT_EARTH_RADIUS),
                ),
                (3, "resolutionAndComponentFlags", 0),
                (3, "latitudeOfSouthernPole", -37500000),
                (3, "longitudeOfSouthernPole", 357500000),
                (3, "angleOfRotation", 0),
            )
            self.assertGribMessageContents(temp_file_path, expect_values)

            # Load the Grib file back into a new cube.
            cube_loaded_from_saved = load_cube(temp_file_path)
            # Also load data, before the temporary file gets deleted.
            cube_loaded_from_saved.data

        # The re-loaded result will not match the original in every respect:
        #  * cube attributes are discarded
        #  * horizontal coordinates are rounded to an integer representation
        #  * bounds on horizontal coords are lost
        # Thus the following "equivalence tests" are rather piecemeal..

        # Check those re-loaded properties which should match the original.
        for test_cube in (cube, cube_loaded_from_saved):
            self.assertEqual(
                test_cube.standard_name, "air_pressure_at_sea_level"
            )
            self.assertEqual(test_cube.units, "Pa")
            self.assertEqual(test_cube.shape, (928, 744))
            self.assertEqual(test_cube.cell_methods, ())

        # Check only the GRIB_PARAM attribute exists on the re-loaded cube.
        # Note: this does *not* match the original, but is as expected.
        self.assertEqual(
            cube_loaded_from_saved.attributes,
            {"GRIB_PARAM": GRIBCode("GRIB2:d000c003n001")},
        )

        # Now remaining to check: coordinates + data...

        # Check they have all the same coordinates.
        co_names = [coord.name() for coord in cube.coords()]
        co_names_reload = [
            coord.name() for coord in cube_loaded_from_saved.coords()
        ]
        self.assertEqual(sorted(co_names_reload), sorted(co_names))

        # Check all the coordinates.
        for coord_name in co_names:
            try:
                co_orig = cube.coord(coord_name)
                co_load = cube_loaded_from_saved.coord(coord_name)

                # Check shape.
                self.assertEqual(
                    co_load.shape,
                    co_orig.shape,
                    'Shape of re-loaded "{}" coord is {} '
                    "instead of {}".format(
                        coord_name, co_load.shape, co_orig.shape
                    ),
                )

                # Check coordinate points equal, within a tolerance.
                self.assertArrayAllClose(
                    co_load.points, co_orig.points, rtol=1.0e-6
                )

                # Check all coords are unbounded.
                # (NOTE: this is not so for the original X and Y coordinates,
                # but Grib does not store those bounds).
                self.assertIsNone(co_load.bounds)

            except AssertionError as err:
                self.assertTrue(
                    False,
                    'Failed on coordinate "{}" : {}'.format(
                        coord_name, str(err)
                    ),
                )

        # Check that main data array also matches.
        self.assertArrayAllClose(cube.data, cube_loaded_from_saved.data)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function
:func:`iris_grib._load_convert.grid_definition_template_4_and_5`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import DimCoord
import numpy as np

from iris_grib._load_convert import (grid_definition_template_4_and_5,
                                     _MDI as MDI)


RESOLUTION = 1e6


class Test(tests.IrisGribTest):
    def setUp(self):
        self.patch('warnings.warn')
        self.patch('iris_grib._load_convert._is_circular', return_value=False)
        self.metadata = {'factories': [], 'references': [],
                         'standard_name': None,
                         'long_name': None, 'units': None, 'attributes': {},
                         'cell_methods': [], 'dim_coords_and_dims': [],
                         'aux_coords_and_dims': []}
        self.cs = mock.sentinel.coord_system
        self.data = np.arange(10, dtype=np.float64)

    def _check(self, section, request_warning,
               expect_warning=False, y_dim=0, x_dim=1):
        this = 'iris_grib._load_convert.options'
        with mock.patch(this, warn_on_unsupported=request_warning):
            metadata = deepcopy(self.metadata)
            # The called being tested.
            grid_definition_template_4_and_5(section, metadata,
                                             'latitude', 'longitude', self.cs)
            expected = deepcopy(self.metadata)
            coord = DimCoord(self.data,
                             standard_name='latitude',
                             units='degrees',
                             coord_system=self.cs)
            expected['dim_coords_and_dims'].append((coord, y_dim))
            coord = DimCoord(self.data,
                             standard_name='longitude',
                             units='degrees',
                             coord_system=self.cs)
            expected['dim_coords_and_dims'].append((coord, x_dim))
            self.assertEqual(metadata, expected)
            if expect_warning:
                self.assertEqual(len(warnings.warn.mock_calls), 1)
                args, kwargs = warnings.warn.call_args
                self.assertIn('resolution and component flags', args[0])
            else:
                self.assertEqual(len(warnings.warn.mock_calls), 0)

    def test_resolution_default_0(self):
        for request_warn in [False, True]:
            section = {'basicAngleOfTheInitialProductionDomain': 0,
                       'subdivisionsOfBasicAngle': 0,
                       'resolutionAndComponentFlags': 0,
                       'longitudes': self.data * RESOLUTION,
                       'latitudes': self.data * RESOLUTION,
                       'scanningMode': 0}
            self._check(section, request_warn)

    def test_resolution_default_mdi(self):
        for request_warn in [False, True]:
            section = {'basicAngleOfTheInitialProductionDomain': MDI,
                       'subdivisionsOfBasicAngle': MDI,
                       'resolutionAndComponentFlags': 0,
                       'longitudes': self.data * RESOLUTION,
                       'latitudes': self.data * RESOLUTION,
                       'scanningMode': 0}
            self._check(section, request_warn)

    def test_resolution(self):
        angle = 10
        for request_warn in [False, True]:
            section = {'basicAngleOfTheInitialProductionDomain': 1,
                       'subdivisionsOfBasicAngle': angle,
                       'resolutionAndComponentFlags': 0,
                       'longitudes': self.data * angle,
                       'latitudes': self.data * angle,
                       'scanningMode': 0}
            self._check(section, request_warn)

    def test_uv_resolved_warn(self):
        angle = 100
        for warn in [False, True]:
            section = {'basicAngleOfTheInitialProductionDomain': 1,
                       'subdivisionsOfBasicAngle': angle,
                       'resolutionAndComponentFlags': 0x08,
                       'longitudes': self.data * angle,
                       'latitudes': self.data * angle,
                       'scanningMode': 0}
            self._check(section, warn, expect_warning=warn)

    def test_j_consecutive(self):
        angle = 1000
        for request_warn in [False, True]:
            section = {'basicAngleOfTheInitialProductionDomain': 1,
                       'subdivisionsOfBasicAngle': angle,
                       'resolutionAndComponentFlags': 0,
                       'longitudes': self.data * angle,
                       'latitudes': self.data * angle,
                       'scanningMode': 0x20}
            self._check(section, request_warn, y_dim=1, x_dim=0)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for `iris_grib._save_rules.fixup_int32_as_uint32`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from iris_grib._save_rules import fixup_int32_as_uint32


class Test(tests.IrisGribTest):
    def test_very_negative(self):
        with self.assertRaises(ValueError):
            fixup_int32_as_uint32(-0x80000000)

    def test_negative(self):
        result = fixup_int32_as_uint32(-3)
        self.assertEqual(result, 0x80000003)

    def test_zero(self):
        result = fixup_int32_as_uint32(0)
        self.assertEqual(result, 0)

    def test_positive(self):
        result = fixup_int32_as_uint32(5)
        self.assertEqual(result, 5)

    def test_very_positive(self):
        with self.assertRaises(ValueError):
            fixup_int32_as_uint32(0x80000000)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris_grib._load_convert` package."""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from collections import OrderedDict


def empty_metadata():
    metadata = OrderedDict()
    metadata['factories'] = []
    metadata['references'] = []
    metadata['standard_name'] = None
    metadata['long_name'] = None
    metadata['units'] = None
    metadata['attributes'] = {}
    metadata['cell_methods'] = []
    metadata['dim_coords_and_dims'] = []
    metadata['aux_coords_and_dims'] = []
    return metadata


class LoadConvertTest(tests.IrisGribTest):
    def assertMetadataEqual(self, result, expected):
        # Compare two metadata dictionaries. Gives slightly more
        # helpful error message than: self.assertEqual(result, expected)
        self.assertEqual(result.keys(), expected.keys())
        for key in result.keys():
            self.assertEqual(result[key], expected[key])



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides testing capabilities for installed copies of iris-grib.

"""

import argparse

from ._runner import TestRunner


parser = argparse.ArgumentParser(
    "iris_grib.tests", description=TestRunner.description
)
for long_opt, short_opt, help_text in TestRunner.user_options:
    long_opt = long_opt.strip("=")
    if long_opt in TestRunner.boolean_options:
        parser.add_argument(
            "--" + long_opt,
            "-" + short_opt,
            action="store_true",
            help=help_text,
        )
    else:
        parser.add_argument("--" + long_opt, "-" + short_opt, help=help_text)
args = parser.parse_args()

runner = TestRunner()

runner.initialize_options()
for long_opt, short_opt, help_text in TestRunner.user_options:
    arg = long_opt.replace("-", "_").strip("=")
    setattr(runner, arg, getattr(args, arg))
runner.finalize_options()

runner.run()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for function
:func:`iris_grib._load_convert.product_definition_template_8`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris_grib._load_convert import product_definition_template_8


class Test(tests.IrisGribTest):
    def setUp(self):
        module = 'iris_grib._load_convert'
        self.module = module
        # Create patches for called routines
        self.patch_generating_process = self.patch(
            module + '.generating_process')
        self.patch_data_cutoff = self.patch(module + '.data_cutoff')
        self.patch_statistical_cell_method = self.patch(
            module + '.statistical_cell_method',
            return_value=mock.sentinel.dummy_cell_method)
        self.patch_statistical_fp_coord = self.patch(
            module + '.statistical_forecast_period_coord',
            return_value=mock.sentinel.dummy_fp_coord)
        self.patch_time_coord = self.patch(
            module + '.validity_time_coord',
            return_value=mock.sentinel.dummy_time_coord)
        self.patch_vertical_coords = self.patch(module + '.vertical_coords')
        # Construct dummy call arguments
        self.section = {}
        self.section['hoursAfterDataCutoff'] = mock.sentinel.cutoff_hours
        self.section['minutesAfterDataCutoff'] = mock.sentinel.cutoff_mins
        self.frt_coord = mock.Mock()
        self.metadata = {'cell_methods': [], 'aux_coords_and_dims': []}

    def test_basic(self):
        product_definition_template_8(
            self.section, self.metadata, self.frt_coord)
        # Check all expected functions were called just once.
        self.assertEqual(self.patch_generating_process.call_count, 1)
        self.assertEqual(self.patch_data_cutoff.call_count, 1)
        self.assertEqual(self.patch_statistical_cell_method.call_count, 1)
        self.assertEqual(self.patch_statistical_fp_coord.call_count, 1)
        self.assertEqual(self.patch_time_coord.call_count, 1)
        self.assertEqual(self.patch_vertical_coords.call_count, 1)
        # Check metadata content.
        self.assertEqual(sorted(self.metadata.keys()),
                         ['aux_coords_and_dims', 'cell_methods'])
        self.assertEqual(self.metadata['cell_methods'],
                         [mock.sentinel.dummy_cell_method])
        self.assertCountEqual(self.metadata['aux_coords_and_dims'],
                              [(self.frt_coord, None),
                               (mock.sentinel.dummy_fp_coord, None),
                               (mock.sentinel.dummy_time_coord, None)])


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules._missing_forecast_period.`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import datetime
import numpy as np

from iris.cube import Cube
from iris.coords import DimCoord

from iris_grib._save_rules import _missing_forecast_period


class TestNoForecastReferenceTime(tests.IrisGribTest):
    def test_no_bounds(self):
        t_coord = DimCoord(15, 'time', units='hours since epoch')
        cube = Cube(23)
        cube.add_aux_coord(t_coord)

        res = _missing_forecast_period(cube)
        expected_rt = t_coord.units.num2date(15)
        expected_rt_type = 3
        expected_fp = 0
        expected_fp_type = 1
        expected = (expected_rt,
                    expected_rt_type,
                    expected_fp,
                    expected_fp_type)
        self.assertEqual(res, expected)

    def test_with_bounds(self):
        t_coord = DimCoord(15, 'time', bounds=[14, 16],
                           units='hours since epoch')
        cube = Cube(23)
        cube.add_aux_coord(t_coord)

        res = _missing_forecast_period(cube)
        expected_rt = t_coord.units.num2date(14)
        expected_rt_type = 3
        expected_fp = 0
        expected_fp_type = 1
        expected = (expected_rt,
                    expected_rt_type,
                    expected_fp,
                    expected_fp_type)
        self.assertEqual(res, expected)


class TestWithForecastReferenceTime(tests.IrisGribTest):
    def test_no_bounds(self):
        t_coord = DimCoord(3, 'time', units='days since epoch')
        frt_coord = DimCoord(8, 'forecast_reference_time',
                             units='hours since epoch')
        cube = Cube(23)
        cube.add_aux_coord(t_coord)
        cube.add_aux_coord(frt_coord)

        res = _missing_forecast_period(cube)
        expected_rt = frt_coord.units.num2date(8)
        expected_rt_type = 1
        expected_fp = 3 * 24 - 8
        expected_fp_type = 1
        expected = (expected_rt,
                    expected_rt_type,
                    expected_fp,
                    expected_fp_type)
        self.assertEqual(res, expected)

    def test_with_bounds(self):
        t_coord = DimCoord(3, 'time', bounds=[2, 4], units='days since epoch')
        frt_coord = DimCoord(8, 'forecast_reference_time',
                             units='hours since epoch')
        cube = Cube(23)
        cube.add_aux_coord(t_coord)
        cube.add_aux_coord(frt_coord)

        res = _missing_forecast_period(cube)
        expected_rt = frt_coord.units.num2date(8)
        expected_rt_type = 1
        expected_fp = 2 * 24 - 8
        expected_fp_type = 1
        expected = (expected_rt,
                    expected_rt_type,
                    expected_fp,
                    expected_fp_type)
        self.assertEqual(res, expected)


if __name__ == "__main__":
    tests.main()



#!/usr/bin/env python

import os
import os.path
from setuptools import setup

NAME = 'iris_grib'
PYPI_NAME = 'iris-grib'
PACKAGE_DIR = os.path.abspath(os.path.dirname(__file__))
PACKAGE_ROOT = os.path.join(PACKAGE_DIR, NAME)

packages = []
for d, _, _ in os.walk(os.path.join(PACKAGE_DIR, NAME)):
    if os.path.exists(os.path.join(d, '__init__.py')):
        packages.append(d[len(PACKAGE_DIR) + 1:].replace(os.path.sep, '.'))


def pip_requirements(*args):
    requirements = []
    for name in args:
        fname = os.path.join(
            PACKAGE_DIR, "requirements", "{}.txt".format(name)
        )
        if not os.path.exists(fname):
            emsg = (
                f"Unable to find the {name!r} requirements file at {fname!r}"
            )
            raise RuntimeError(emsg)
        with open(fname, "r") as fh:
            for line in fh:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                requirements.append(line)
    return requirements


def extract_version():
    version = None
    fname = os.path.join(PACKAGE_DIR, 'iris_grib', '__init__.py')
    with open(fname) as fi:
        for line in fi:
            if (line.startswith('__version__')):
                _, version = line.split('=')
                version = version.strip()[1:-1]  # Remove quotations
                break
    return version


def long_description():
    fname = os.path.join(PACKAGE_DIR, "README.rst")
    with open(fname, "rb") as fi:
        result = fi.read().decode("utf-8")
    return result


def file_walk_relative(top, remove=''):
    """
    Returns a generator of files from the top of the tree, removing
    the given prefix from the root/file result.

    """
    top = top.replace('/', os.path.sep)
    remove = remove.replace('/', os.path.sep)
    for root, dirs, files in os.walk(top):
        for file in files:
            yield os.path.join(root, file).replace(remove, '')


setup_args = dict(
    name             = PYPI_NAME,
    version          = extract_version(),
    packages         = packages,
    package_data     = {'iris_grib': list(file_walk_relative('iris_grib/tests/results',
                                          remove='iris_grib/'))},
    description      = "GRIB loading for Iris",
    long_description = long_description(),
    long_description_content_type = "text/x-rst",
    url              = 'https://github.com/SciTools/iris-grib',
    author           = 'UK Met Office',
    author_email     = 'scitools-iris@googlegroups.com',
    license          = 'LGPL',
    platforms        = "Linux, Mac OS X, Windows",
    keywords         = ['iris', 'GRIB'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    # NOTE: The Python 3 bindings to eccodes (eccodes-python) is available on
    # PyPI, but the user is required to install eccodes itself manually. See
    # ECMWF ecCodes installation documentation for more information.
    install_requires=pip_requirements("setup", "core"),
    test_loader="unittest:TestLoader",
    extras_require={
        "all": pip_requirements("all"),
        "test": pip_requirements("test"),
    },
)


if __name__ == '__main__':
    setup(**setup_args)



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.ellipsoid.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy.ma as ma

import iris.coord_systems as icoord_systems
from iris.exceptions import TranslationError

from iris_grib._load_convert import ellipsoid


# Reference GRIB2 Code Table 3.2 - Shape of the Earth.


MDI = ma.masked


class Test(tests.IrisGribTest):
    def test_shape_unsupported(self):
        unsupported = [8, 9, 10, MDI]
        emsg = 'unsupported shape of the earth'
        for shape in unsupported:
            with self.assertRaisesRegex(TranslationError, emsg):
                ellipsoid(shape, MDI, MDI, MDI)

    def test_spherical_default_supported(self):
        cs_by_shape = {0: icoord_systems.GeogCS(6367470),
                       6: icoord_systems.GeogCS(6371229)}
        for shape, expected in cs_by_shape.items():
            result = ellipsoid(shape, MDI, MDI, MDI)
            self.assertEqual(result, expected)

    def test_spherical_shape_1_no_radius(self):
        shape = 1
        emsg = 'radius to be specified'
        with self.assertRaisesRegex(ValueError, emsg):
            ellipsoid(shape, MDI, MDI, MDI)

    def test_spherical_shape_1(self):
        shape = 1
        radius = 10
        result = ellipsoid(shape, MDI, MDI, radius)
        expected = icoord_systems.GeogCS(radius)
        self.assertEqual(result, expected)

    def test_oblate_shape_3_7_no_axes(self):
        for shape in [3, 7]:
            emsg = 'axis to be specified'
            with self.assertRaisesRegex(ValueError, emsg):
                ellipsoid(shape, MDI, MDI, MDI)

    def test_oblate_shape_3_7_no_major(self):
        for shape in [3, 7]:
            emsg = 'major axis to be specified'
            with self.assertRaisesRegex(ValueError, emsg):
                ellipsoid(shape, MDI, 1, MDI)

    def test_oblate_shape_3_7_no_minor(self):
        for shape in [3, 7]:
            emsg = 'minor axis to be specified'
            with self.assertRaisesRegex(ValueError, emsg):
                ellipsoid(shape, 1, MDI, MDI)

    def test_oblate_shape_3_7(self):
        for shape in [3, 7]:
            major, minor = 1, 10
            scale = 1
            result = ellipsoid(shape, major, minor, MDI)
            if shape == 3:
                # Convert km to m.
                scale = 1000
            expected = icoord_systems.GeogCS(major * scale, minor * scale)
            self.assertEqual(result, expected)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris_grib.save_grib2` function."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import iris_grib


class TestSaveGrib2(tests.IrisGribTest):
    def setUp(self):
        self.cube = mock.sentinel.cube
        self.target = mock.sentinel.target
        func = 'iris_grib.save_pairs_from_cube'
        self.messages = list(range(10))
        slices = self.messages
        side_effect = [zip(slices, self.messages)]
        self.save_pairs_from_cube = self.patch(func, side_effect=side_effect)
        func = 'iris_grib.save_messages'
        self.save_messages = self.patch(func)

    def _check(self, append=False):
        iris_grib.save_grib2(self.cube, self.target, append=append)
        self.save_pairs_from_cube.called_once_with(self.cube)
        args, kwargs = self.save_messages.call_args
        self.assertEqual(len(args), 2)
        messages, target = args
        self.assertEqual(list(messages), self.messages)
        self.assertEqual(target, self.target)
        self.assertEqual(kwargs, dict(append=append))

    def test_save_no_append(self):
        self._check()

    def test_save_append(self):
        self._check(append=True)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris_grib._load_convert.time_range_unit.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from cf_units import Unit
from iris.exceptions import TranslationError

from iris_grib._load_convert import time_range_unit


class Test(tests.IrisGribTest):
    def setUp(self):
        self.unit_by_indicator = {0: Unit('minutes'),
                                  1: Unit('hours'),
                                  2: Unit('days'),
                                  10: Unit('3 hours'),
                                  11: Unit('6 hours'),
                                  12: Unit('12 hours'),
                                  13: Unit('seconds')}

    def test_units(self):
        for indicator, unit in self.unit_by_indicator.items():
            result = time_range_unit(indicator)
            self.assertEqual(result, unit)

    def test_bad_indicator(self):
        emsg = 'unsupported time range'
        with self.assertRaisesRegex(TranslationError, emsg):
            time_range_unit(-1)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.product_definition_template_15`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

from cf_units import Unit
import gribapi

from iris.coords import CellMethod, DimCoord
import iris.tests.stock as stock

from iris_grib._save_rules import product_definition_template_15


class TestSpatialProcessingIdentifiers(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Add scalar time coord so that product_definition_template_common
        # doesn't get upset.
        t_coord = DimCoord([424854.], standard_name='time',
                           units=Unit('hours since 1970-01-01 00:00:00',
                                      calendar='gregorian'))
        self.cube.add_aux_coord(t_coord)
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename('air_temperature')

    @mock.patch.object(gribapi, 'grib_set')
    def test_cell_method(self, mock_set):
        cube_0 = self.cube
        cube_0.attributes = dict(spatial_processing_type='No interpolation')
        cell_method = CellMethod(method='mean', coords=['area'])
        cube_0.add_cell_method(cell_method)

        # If the cube has a cell method attached then it should not have any
        # interpolation on the data, so spatial processing code should be 0 and
        # number of points used should be 0.
        product_definition_template_15(cube_0, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 15)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "spatialProcessing", 0)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "statisticalProcess", 0)

    @mock.patch.object(gribapi, 'grib_set')
    def test_bilinear_interpolation(self, mock_set):
        cube_1 = self.cube
        cube_1.attributes = dict(spatial_processing_type='Bilinear '
                                                         'interpolation')

        # If the cube has a bilinear interpolation attribute, spatial
        # processing code should be 1 and number of points used should be 4.
        product_definition_template_15(cube_1, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 15)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "spatialProcessing", 1)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "numberOfPointsUsed", 4)

    @mock.patch.object(gribapi, 'grib_set')
    def test_bicubic_interpolation(self, mock_set):
        cube_2 = self.cube
        cube_2.attributes = dict(spatial_processing_type='Bicubic '
                                                         'interpolation')

        # If the cube has a bicubic interpolation attribute, spatial
        # processing code should be 2 and number of points used should be 4.
        product_definition_template_15(cube_2, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 15)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "spatialProcessing", 2)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "numberOfPointsUsed", 4)

    @mock.patch.object(gribapi, 'grib_set')
    def test_nearest_neighbour_interpolation(self, mock_set):
        cube_3 = self.cube
        cube_3.attributes = dict(spatial_processing_type='Nearest neighbour '
                                                         'interpolation')

        # If the cube has a nearest neighbour interpolation attribute, spatial
        # processing code should be 3 and number of points used should be 1.
        product_definition_template_15(cube_3, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 15)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "spatialProcessing", 3)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "numberOfPointsUsed", 1)

    @mock.patch.object(gribapi, 'grib_set')
    def test_budget_interpolation(self, mock_set):
        cube_4 = self.cube
        cube_4.attributes = dict(spatial_processing_type='Budget '
                                                         'interpolation')

        # If the cube has a budget interpolation attribute, spatial
        # processing code should be 4 and number of points used should be 4.
        product_definition_template_15(cube_4, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 15)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "spatialProcessing", 4)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "numberOfPointsUsed", 4)

    @mock.patch.object(gribapi, 'grib_set')
    def test_spectral_interpolation(self, mock_set):
        cube_5 = self.cube
        cube_5.attributes = dict(spatial_processing_type='Spectral '
                                                         'interpolation')

        # If the cube has a spectral interpolation attribute, spatial
        # processing code should be 5 and number of points used should be 4.
        product_definition_template_15(cube_5, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 15)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "spatialProcessing", 5)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "numberOfPointsUsed", 4)

    @mock.patch.object(gribapi, 'grib_set')
    def test_neighbour_budget_interpolation(self, mock_set):
        cube_6 = self.cube
        cube_6.attributes = dict(spatial_processing_type='Neighbour-budget '
                                                         'interpolation')

        # If the cube has a neighbour-budget interpolation attribute, spatial
        # processing code should be 6 and number of points used should be 4.
        product_definition_template_15(cube_6, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 15)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "spatialProcessing", 6)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "numberOfPointsUsed", 4)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for `iris_grib._load_convert.product_definition_section`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from itertools import product
from unittest import mock

from iris.coords import DimCoord

from iris_grib._load_convert import product_definition_section
from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib.tests.unit.load_convert.test_product_definition_template_0 \
    import section_4 as pdt_0_section_4
from iris_grib.tests.unit.load_convert.test_product_definition_template_31 \
    import section_4 as pdt_31_section_4


class TestFixedSurfaces(tests.IrisGribTest):
    """
    Tests focussing on the handling of fixed surface elements in section 4.
    Expects/ignores depending on the template number.
    """
    def setUp(self):
        self.patch('warnings.warn')
        self.translate_phenomenon_patch = self.patch(
            'iris_grib._load_convert.translate_phenomenon'
        )

        # Prep placeholder variables for product_definition_section.
        self.discipline = mock.sentinel.discipline
        self.tablesVersion = mock.sentinel.tablesVersion
        self.rt_coord = DimCoord(24, 'forecast_reference_time',
                                 units='hours since epoch')
        self.metadata = empty_metadata()

        self.templates = {0: pdt_0_section_4(), 31: pdt_31_section_4()}
        self.fixed_surface_keys = [
            'typeOfFirstFixedSurface',
            'scaledValueOfFirstFixedSurface',
            'typeOfSecondFixedSurface'
        ]

    def _check_fixed_surface(self, fs_is_expected, fs_is_present):
        """
        Whether or not fixed surface elements are expected/present in the
        section 4 keys, most of the code is shared so we are using a single
        function with parameters.
        """

        # Use the section 4 from either product_definition_section #1 or #31.
        # #0 contains fixed surface elements, #31 does not.
        template_number = 0 if fs_is_expected else 31
        section_4 = self.templates[template_number]
        section_4.update({
            'productDefinitionTemplateNumber': template_number,
            'parameterCategory': None,
            'parameterNumber': None
        })

        for key in self.fixed_surface_keys:
            # Force the presence or absence of the fixed surface elements even
            # when they're respectively ignored or expected.
            if fs_is_present and key not in section_4:
                section_4[key] = pdt_0_section_4()[key]
            elif (not fs_is_present) and key in section_4:
                del section_4[key]

        def run_function():
            # For re-use in every type of test below.
            product_definition_section(
                section_4, self.metadata, self.discipline, self.tablesVersion,
                self.rt_coord)

        if fs_is_expected and not fs_is_present:
            # Should error since the expected keys are missing.
            error_message = 'FixedSurface'
            with self.assertRaisesRegex(KeyError, error_message):
                run_function()
        else:
            # Should have a successful run for all other circumstances.

            # Translate_phenomenon_patch is the end of the function,
            # and should be able to accept None for the fixed surface
            # arguments. So should always have run.
            previous_call_count = self.translate_phenomenon_patch.call_count
            run_function()
            self.assertEqual(self.translate_phenomenon_patch.call_count,
                             previous_call_count + 1)
            phenom_call_args = self.translate_phenomenon_patch.call_args[1]
            for key in self.fixed_surface_keys:
                # Check whether None or actual values have been passed for
                # the fixed surface arguments.
                if fs_is_expected:
                    self.assertEqual(phenom_call_args[key], section_4[key])
                else:
                    self.assertIsNone(phenom_call_args[key])

    def test_all_combinations(self):
        """
        Test all combinations of fixed surface being expected/present

        a. Expected and Present - standard behaviour for most templates
        b. Expected and Absent - unplanned combination, should error
        c. Unexpected and Present - unplanned combination, should be handled
            identically to (d)
        d. Unexpected and Absent - standard behaviour for a few templates
            e.g. #31
        """
        for pair in product([True, False], repeat=2):
            self._check_fixed_surface(*pair)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris_grib._load_generate` function."""

import iris_grib.tests as tests

from unittest import mock

import iris
from iris.exceptions import TranslationError
from iris.fileformats.rules import Loader

import iris_grib
from iris_grib import GribWrapper
from iris_grib import _load_generate
from iris_grib.message import GribMessage


class Test(tests.IrisGribTest):
    def setUp(self):
        self.fname = mock.sentinel.fname
        self.message_id = mock.sentinel.message_id
        self.grib_fh = mock.sentinel.grib_fh

    def _make_test_message(self, sections):
        raw_message = mock.Mock(sections=sections, _message_id=self.message_id)
        file_ref = mock.Mock(open_file=self.grib_fh)
        return GribMessage(raw_message, None, file_ref=file_ref)

    def test_grib1(self):
        sections = [{'editionNumber': 1}]
        message = self._make_test_message(sections)
        mfunc = 'iris_grib.GribMessage.messages_from_filename'
        mclass = 'iris_grib.GribWrapper'
        with mock.patch(mfunc, return_value=[message]) as mock_func:
            with mock.patch(mclass, spec=GribWrapper) as mock_wrapper:
                field = next(_load_generate(self.fname))
                mock_func.assert_called_once_with(self.fname)
                self.assertIsInstance(field, GribWrapper)
                mock_wrapper.assert_called_once_with(self.message_id,
                                                     grib_fh=self.grib_fh)

    def test_grib2(self):
        sections = [{'editionNumber': 2}]
        message = self._make_test_message(sections)
        mfunc = 'iris_grib.GribMessage.messages_from_filename'
        with mock.patch(mfunc, return_value=[message]) as mock_func:
            field = next(_load_generate(self.fname))
            mock_func.assert_called_once_with(self.fname)
            self.assertEqual(field, message)

    def test_grib_unknown(self):
        sections = [{'editionNumber': 0}]
        message = self._make_test_message(sections)
        mfunc = 'iris_grib.GribMessage.messages_from_filename'
        emsg = 'GRIB edition 0 is not supported'
        with mock.patch(mfunc, return_value=[message]):
            with self.assertRaisesRegex(TranslationError, emsg):
                next(_load_generate(self.fname))


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration test for round-trip loading and saving of hybrid height and
hybrid pressure cubes.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris import load_cube, load, save
from iris.cube import Cube


@tests.skip_grib_data
class TestWAFCCodes(tests.IrisGribTest):
    def setUp(self):
        self.cat = self.get_testdata_path('CAT_T+24_0600.grib2')
        self.cb = self.get_testdata_path('CB_T+24_0600.grib2')
        self.icing = self.get_testdata_path('ICING_T+24_0600.grib2')
        self.turb = self.get_testdata_path('INCLDTURB_T+24_0600.grib2')

    def test_WAFC_CAT_round_trip(self):
        cubelist = load(self.cat, 'WAFC_CAT_potential')
        cube = cubelist[0]
        self.assertIsInstance(cube, Cube)

        with self.temp_filename() as tmp_save_path:
            save(cube, tmp_save_path, saver='grib2')
            saved_cube = load_cube(tmp_save_path)
            self.assertEqual(saved_cube.metadata, cube.metadata)

    def test_WAFC_CB_round_trip(self):
        cubelist = load(self.cb)
        cube = cubelist[0]
        self.assertIsInstance(cube, Cube)

        with self.temp_filename() as tmp_save_path:
            save(cube, tmp_save_path, saver='grib2')
            saved_cube = load_cube(tmp_save_path)
            self.assertEqual(saved_cube.metadata, cube.metadata)

    def test_WAFC_icing_round_trip(self):
        cubelist = load(self.icing)
        cube = cubelist[0]
        self.assertIsInstance(cube, Cube)

        with self.temp_filename() as tmp_save_path:
            save(cube, tmp_save_path, saver='grib2')
            saved_cube = load_cube(tmp_save_path)
            self.assertEqual(saved_cube.metadata, cube.metadata)

    def test_WAFC_turb_round_trip(self):
        cubelist = load(self.turb)
        cube = cubelist[0]
        self.assertIsInstance(cube, Cube)

        with self.temp_filename() as tmp_save_path:
            save(cube, tmp_save_path, saver='grib2')
            saved_cube = load_cube(tmp_save_path)
            self.assertEqual(saved_cube.metadata, cube.metadata)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.product_definition_template_8`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

from cf_units import Unit
import gribapi
import numpy as np

from iris.coords import CellMethod, DimCoord
import iris.cube
import iris.tests.stock as stock

from iris_grib._save_rules import product_definition_template_8


class TestProductDefinitionIdentifier(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename('air_temperature')
        coord = DimCoord(23, 'time', bounds=[0, 100],
                         units=Unit('days since epoch', calendar='standard'))
        self.cube.add_aux_coord(coord)

    @mock.patch.object(gribapi, 'grib_set')
    def test_product_definition(self, mock_set):
        cube = self.cube
        cell_method = CellMethod(method='sum', coords=['time'])
        cube.add_cell_method(cell_method)

        product_definition_template_8(cube, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib,
                                 "productDefinitionTemplateNumber", 8)


class Test_type_of_statistical_processing(tests.IrisTest):
    @mock.patch.object(gribapi, "grib_set")
    def test_stats_type_min(self, mock_set):
        grib = None
        cube = iris.cube.Cube(np.array([1.0]))
        time_unit = Unit("hours since 1970-01-01 00:00:00")
        time_coord = iris.coords.DimCoord(
            [0.0], bounds=[0.0, 1], standard_name="time", units=time_unit
        )
        cube.add_aux_coord(time_coord, ())
        cube.add_cell_method(iris.coords.CellMethod("maximum", time_coord))
        product_definition_template_8(cube, grib)
        mock_set.assert_any_call(grib, "typeOfStatisticalProcessing", 2)

    @mock.patch.object(gribapi, "grib_set")
    def test_stats_type_max(self, mock_set):
        grib = None
        cube = iris.cube.Cube(np.array([1.0]))
        time_unit = Unit("hours since 1970-01-01 00:00:00")
        time_coord = iris.coords.DimCoord(
            [0.0], bounds=[0.0, 1], standard_name="time", units=time_unit
        )
        cube.add_aux_coord(time_coord, ())
        cube.add_cell_method(iris.coords.CellMethod("minimum", time_coord))
        product_definition_template_8(cube, grib)
        mock_set.assert_any_call(grib, "typeOfStatisticalProcessing", 3)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_90`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords
import iris.exceptions

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_90


class Test(tests.IrisGribTest):
    def uk(self):
        section = {
            'shapeOfTheEarth': 3,
            'scaleFactorOfRadiusOfSphericalEarth': MDI,
            'scaledValueOfRadiusOfSphericalEarth': MDI,
            'scaleFactorOfEarthMajorAxis': 4,
            'scaledValueOfEarthMajorAxis': 63781688,
            'scaleFactorOfEarthMinorAxis': 4,
            'scaledValueOfEarthMinorAxis': 63565840,
            'Nx': 390,
            'Ny': 227,
            'latitudeOfSubSatellitePoint': 0,
            'longitudeOfSubSatellitePoint': 0,
            'resolutionAndComponentFlags': 0,
            'dx': 3622,
            'dy': 3610,
            'Xp': 1856000,
            'Yp': 1856000,
            'scanningMode': 192,
            'orientationOfTheGrid': 0,
            'Nr': 6610674,
            'Xo': 1733,
            'Yo': 3320
        }
        return section

    def expected_uk(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        major = 6378168.8
        ellipsoid = iris.coord_systems.GeogCS(major, 6356584.0)
        height = (6610674e-6 - 1) * major
        lat = lon = 0
        easting = northing = 0
        cs = iris.coord_systems.Geostationary(
            latitude_of_projection_origin=lat,
            longitude_of_projection_origin=lon,
            perspective_point_height=height,
            sweep_angle_axis='y',
            false_easting=easting,
            false_northing=northing,
            ellipsoid=ellipsoid)
        nx = 390
        x_origin = 0.010313624253429191
        dx = -8.38506036864162e-05
        x = iris.coords.DimCoord(np.arange(nx) * dx + x_origin,
                                 'projection_x_coordinate', units='radians',
                                 coord_system=cs)
        ny = 227
        y_origin = 0.12275487535118533
        dy = 8.384895857321404e-05
        y = iris.coords.DimCoord(np.arange(ny) * dy + y_origin,
                                 'projection_y_coordinate', units='radians',
                                 coord_system=cs)
        expected['dim_coords_and_dims'].append((y, y_dim))
        expected['dim_coords_and_dims'].append((x, x_dim))
        return expected

    def compare(self, metadata, expected):
        # Compare the result with the expectation.
        self.assertEqual(len(metadata['dim_coords_and_dims']),
                         len(expected['dim_coords_and_dims']))
        for result_pair, expected_pair in zip(metadata['dim_coords_and_dims'],
                                              expected['dim_coords_and_dims']):
            result_coord, result_dims = result_pair
            expected_coord, expected_dims = expected_pair
            # Take copies for safety, as we are going to modify them.
            result_coord, expected_coord = [co.copy() for co in
                                            (result_coord, expected_coord)]
            # Ensure the dims match.
            self.assertEqual(result_dims, expected_dims)
            # Ensure the coordinate systems match (allowing for precision).
            result_cs = result_coord.coord_system
            expected_cs = expected_coord.coord_system
            self.assertEqual(type(result_cs), type(expected_cs))
            self.assertEqual(result_cs.latitude_of_projection_origin,
                             expected_cs.latitude_of_projection_origin)
            self.assertEqual(result_cs.longitude_of_projection_origin,
                             expected_cs.longitude_of_projection_origin)
            self.assertAlmostEqual(result_cs.perspective_point_height,
                                   expected_cs.perspective_point_height)
            self.assertEqual(result_cs.false_easting,
                             expected_cs.false_easting)
            self.assertEqual(result_cs.false_northing,
                             expected_cs.false_northing)
            self.assertAlmostEqual(result_cs.ellipsoid.semi_major_axis,
                                   expected_cs.ellipsoid.semi_major_axis)
            self.assertEqual(result_cs.ellipsoid.semi_minor_axis,
                             expected_cs.ellipsoid.semi_minor_axis)
            # Now we can ignore the coordinate systems and compare the
            # rest of the coordinate attributes.
            result_coord.coord_system = None
            expected_coord.coord_system = None

            # Likewise, first compare the points (and optional bounds)
            # *approximately*, then force those equal + compare other aspects.
            self.assertArrayAlmostEqual(result_coord.points,
                                        expected_coord.points)
            result_coord.points = expected_coord.points
            if result_coord.has_bounds() and expected_coord.has_bounds():
                self.assertArrayAlmostEqual(result_coord.bounds,
                                            expected_coord.bounds)
                result_coord.bounds = expected_coord.bounds

            # Compare the coords, having equalised the array values.
            self.assertEqual(result_coord, expected_coord)

        # Ensure no other metadata was created.
        for name in expected.keys():
            if name == 'dim_coords_and_dims':
                continue
            self.assertEqual(metadata[name], expected[name])

    def test_uk(self):
        section = self.uk()
        metadata = empty_metadata()
        grid_definition_template_90(section, metadata)
        expected = self.expected_uk(0, 1)
        self.compare(metadata, expected)

    def test_uk_transposed(self):
        section = self.uk()
        section['scanningMode'] = 0b11100000
        metadata = empty_metadata()
        grid_definition_template_90(section, metadata)
        expected = self.expected_uk(1, 0)
        self.compare(metadata, expected)

    def test_non_zero_latitude(self):
        section = self.uk()
        section['latitudeOfSubSatellitePoint'] = 1
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError,
                                    'non-zero latitude'):
            grid_definition_template_90(section, metadata)

    def test_rotated_meridian(self):
        section = self.uk()
        section['orientationOfTheGrid'] = 1
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError,
                                    'orientation'):
            grid_definition_template_90(section, metadata)

    def test_zero_height(self):
        section = self.uk()
        section['Nr'] = 0
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError,
                                    'zero'):
            grid_definition_template_90(section, metadata)

    def test_orthographic(self):
        section = self.uk()
        section['Nr'] = MDI
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError,
                                    'orthographic'):
            grid_definition_template_90(section, metadata)

    def test_scanning_mode_positive_x(self):
        section = self.uk()
        section['scanningMode'] = 0b01000000
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError, r'\+x'):
            grid_definition_template_90(section, metadata)

    def test_scanning_mode_negative_y(self):
        section = self.uk()
        section['scanningMode'] = 0b10000000
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError, '-y'):
            grid_definition_template_90(section, metadata)


if __name__ == '__main__':
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Conversion of cubes to/from GRIB.

See: `ECMWF GRIB API <https://software.ecmwf.int/wiki/display/GRIB/Home>`_.

"""

import datetime
import math  # for fmod
import warnings

import cartopy.crs as ccrs
import cf_units
import gribapi
import numpy as np
import numpy.ma as ma

# NOTE: careful here, to avoid circular imports (as iris imports grib)
import iris
from iris._lazy_data import as_lazy_data
import iris.coord_systems as coord_systems
from iris.exceptions import TranslationError, NotYetImplementedError

from . import grib_phenom_translation as gptx
from . import _save_rules
from ._load_convert import convert as load_convert
from .message import GribMessage


__version__ = '0.17.1'

__all__ = ['load_cubes', 'save_grib2', 'load_pairs_from_fields',
           'save_pairs_from_cube', 'save_messages']


CENTRE_TITLES = {'egrr': 'U.K. Met Office - Exeter',
                 'ecmf': 'European Centre for Medium Range Weather Forecasts',
                 'rjtd': 'Tokyo, Japan Meteorological Agency',
                 '55': 'San Francisco',
                 'kwbc': ('US National Weather Service, National Centres for '
                          'Environmental Prediction')}

TIME_RANGE_INDICATORS = {0: 'none', 1: 'none', 3: 'time mean', 4: 'time sum',
                         5: 'time _difference', 10: 'none',
                         # TODO #567 Further exploration of following mappings
                         51: 'time mean', 113: 'time mean', 114: 'time sum',
                         115: 'time mean', 116: 'time sum', 117: 'time mean',
                         118: 'time _covariance', 123: 'time mean',
                         124: 'time sum', 125: 'time standard_deviation'}

PROCESSING_TYPES = {0: 'time mean', 1: 'time sum', 2: 'time maximum',
                    3: 'time minimum', 4: 'time _difference',
                    5: 'time _root mean square', 6: 'time standard_deviation',
                    7: 'time _convariance', 8: 'time _difference',
                    9: 'time _ratio'}

TIME_CODES_EDITION1 = {
    0: ('minutes', 60),
    1: ('hours', 60*60),
    2: ('days', 24*60*60),
    # NOTE: do *not* support calendar-dependent units at all.
    # So the following possible keys remain unsupported:
    #  3: 'months',
    #  4: 'years',
    #  5: 'decades',
    #  6: '30 years',
    #  7: 'century',
    10: ('3 hours', 3*60*60),
    11: ('6 hours', 6*60*60),
    12: ('12 hours', 12*60*60),
    13: ('15 minutes', 15*60),
    14: ('30 minutes', 30*60),
    254: ('seconds', 1),
}

unknown_string = "???"


class GribDataProxy:
    """A reference to the data payload of a single Grib message."""

    __slots__ = ('shape', 'dtype', 'path', 'offset')

    def __init__(self, shape, dtype, path, offset):
        self.shape = shape
        self.dtype = dtype
        self.path = path
        self.offset = offset

    @property
    def ndim(self):
        return len(self.shape)

    def __getitem__(self, keys):
        with open(self.path, 'rb') as grib_fh:
            grib_fh.seek(self.offset)
            grib_message = gribapi.grib_new_from_file(grib_fh)
            data = _message_values(grib_message, self.shape)
            gribapi.grib_release(grib_message)

        return data.__getitem__(keys)

    def __repr__(self):
        msg = '<{self.__class__.__name__} shape={self.shape} ' \
            'dtype={self.dtype!r} ' \
            'path={self.path!r} offset={self.offset}>'
        return msg.format(self=self)

    def __getstate__(self):
        return {attr: getattr(self, attr) for attr in self.__slots__}

    def __setstate__(self, state):
        for key, value in state.items():
            setattr(self, key, value)


class GribWrapper:
    """
    Contains a pygrib object plus some extra keys of our own.

    The class :class:`iris_grib.message.GribMessage`
    provides alternative means of working with GRIB message instances.

    """
    def __init__(self, grib_message, grib_fh=None):
        """Store the grib message and compute our extra keys."""
        self.grib_message = grib_message

        if self.edition != 1:
            emsg = 'GRIB edition {} is not supported by {!r}.'
            raise TranslationError(emsg.format(self.edition,
                                               type(self).__name__))

        deferred = grib_fh is not None

        # Store the file pointer and message length from the current
        # grib message before it's changed by calls to the grib-api.
        if deferred:
            # Note that, the grib-api has already read this message and
            # advanced the file pointer to the end of the message.
            offset = grib_fh.tell()
            message_length = gribapi.grib_get_long(grib_message, 'totalLength')

        # Initialise the key-extension dictionary.
        # NOTE: this attribute *must* exist, or the the __getattr__ overload
        # can hit an infinite loop.
        self.extra_keys = {}
        self._confirm_in_scope()
        self._compute_extra_keys()

        # Calculate the data payload shape.
        shape = (gribapi.grib_get_long(grib_message, 'numberOfValues'),)

        if not self.gridType.startswith('reduced'):
            ni, nj = self.Ni, self.Nj
            j_fast = gribapi.grib_get_long(grib_message,
                                           'jPointsAreConsecutive')
            shape = (nj, ni) if j_fast == 0 else (ni, nj)

        if deferred:
            # Wrap the reference to the data payload within the data proxy
            # in order to support deferred data loading.
            # The byte offset requires to be reset back to the first byte
            # of this message. The file pointer offset is always at the end
            # of the current message due to the grib-api reading the message.
            proxy = GribDataProxy(shape, np.array([0.]).dtype, grib_fh.name,
                                  offset - message_length)
            self._data = as_lazy_data(proxy)
        else:
            self.data = _message_values(grib_message, shape)

    def _confirm_in_scope(self):
        """Ensure we have a grib flavour that we choose to support."""

        # forbid alternate row scanning
        # (uncommon entry from GRIB2 flag table 3.4, also in GRIB1)
        if self.alternativeRowScanning == 1:
            raise ValueError("alternativeRowScanning == 1 not handled.")

    def __getattr__(self, key):
        """Return a grib key, or one of our extra keys."""

        # is it in the grib message?
        try:
            # we just get <type 'float'> as the type of the "values"
            # array...special case here...
            if key in ["values", "pv", "latitudes", "longitudes"]:
                res = gribapi.grib_get_double_array(self.grib_message, key)
            elif key in ('typeOfFirstFixedSurface',
                         'typeOfSecondFixedSurface'):
                res = np.int32(gribapi.grib_get_long(self.grib_message, key))
            else:
                key_type = gribapi.grib_get_native_type(self.grib_message, key)
                if key_type == int:
                    res = np.int32(gribapi.grib_get_long(self.grib_message,
                                                         key))
                elif key_type == float:
                    # Because some computer keys are floats, like
                    # longitudeOfFirstGridPointInDegrees, a float32
                    # is not always enough...
                    res = np.float64(gribapi.grib_get_double(self.grib_message,
                                                             key))
                elif key_type == str:
                    res = gribapi.grib_get_string(self.grib_message, key)
                else:
                    emsg = "Unknown type for {} : {}"
                    raise ValueError(emsg.format(key, str(key_type)))
        except gribapi.errors.GribInternalError:
            res = None

        # ...or is it in our list of extras?
        if res is None:
            if key in self.extra_keys:
                res = self.extra_keys[key]
            else:
                # must raise an exception for the hasattr() mechanism to work
                raise AttributeError("Cannot find GRIB key %s" % key)

        return res

    def _timeunit_detail(self):
        """Return the (string, seconds) describing the message time unit."""
        unit_code = self.indicatorOfUnitOfTimeRange
        if unit_code not in TIME_CODES_EDITION1:
            message = 'Unhandled time unit for forecast ' \
                      'indicatorOfUnitOfTimeRange : ' + str(unit_code)
            raise NotYetImplementedError(message)
        return TIME_CODES_EDITION1[unit_code]

    def _timeunit_string(self):
        """Get the udunits string for the message time unit."""
        return self._timeunit_detail()[0]

    def _timeunit_seconds(self):
        """Get the number of seconds in the message time unit."""
        return self._timeunit_detail()[1]

    def _compute_extra_keys(self):
        """Compute our extra keys."""
        global unknown_string

        self.extra_keys = {}
        forecastTime = self.startStep

        # regular or rotated grid?
        try:
            longitudeOfSouthernPoleInDegrees = \
                self.longitudeOfSouthernPoleInDegrees
            latitudeOfSouthernPoleInDegrees = \
                self.latitudeOfSouthernPoleInDegrees
        except AttributeError:
            longitudeOfSouthernPoleInDegrees = 0.0
            latitudeOfSouthernPoleInDegrees = 90.0

        centre = gribapi.grib_get_string(self.grib_message, "centre")

        # default values
        self.extra_keys = {'_referenceDateTime': -1.0,
                           '_phenomenonDateTime': -1.0,
                           '_periodStartDateTime': -1.0,
                           '_periodEndDateTime': -1.0,
                           '_levelTypeName': unknown_string,
                           '_levelTypeUnits': unknown_string,
                           '_firstLevelTypeName': unknown_string,
                           '_firstLevelTypeUnits': unknown_string,
                           '_firstLevel': -1.0,
                           '_secondLevelTypeName': unknown_string,
                           '_secondLevel': -1.0,
                           '_originatingCentre': unknown_string,
                           '_forecastTime': None,
                           '_forecastTimeUnit': unknown_string,
                           '_coord_system': None,
                           '_x_circular': False,
                           '_x_coord_name': unknown_string,
                           '_y_coord_name': unknown_string,
                           # These are here to avoid repetition in the rules
                           # files, and reduce the very long line lengths.
                           '_x_points': None,
                           '_y_points': None,
                           '_cf_data': None}

        # cf phenomenon translation
        # Get centre code (N.B. self.centre has default type = string)
        centre_number = gribapi.grib_get_long(self.grib_message, "centre")
        # Look for a known grib1-to-cf translation (or None).
        cf_data = gptx.grib1_phenom_to_cf_info(
            table2_version=self.table2Version,
            centre_number=centre_number,
            param_number=self.indicatorOfParameter)
        self.extra_keys['_cf_data'] = cf_data

        # reference date
        self.extra_keys['_referenceDateTime'] = \
            datetime.datetime(int(self.year), int(self.month), int(self.day),
                              int(self.hour), int(self.minute))

        # forecast time with workarounds
        self.extra_keys['_forecastTime'] = forecastTime

        # verification date
        processingDone = self._get_processing_done()
        # time processed?
        if processingDone.startswith("time"):
            validityDate = str(self.validityDate)
            validityTime = "{:04}".format(int(self.validityTime))
            endYear = int(validityDate[:4])
            endMonth = int(validityDate[4:6])
            endDay = int(validityDate[6:8])
            endHour = int(validityTime[:2])
            endMinute = int(validityTime[2:4])

            # fixed forecastTime in hours
            self.extra_keys['_periodStartDateTime'] = \
                (self.extra_keys['_referenceDateTime'] +
                 datetime.timedelta(hours=int(forecastTime)))
            self.extra_keys['_periodEndDateTime'] = \
                datetime.datetime(endYear, endMonth, endDay, endHour,
                                  endMinute)
        else:
            self.extra_keys['_phenomenonDateTime'] = \
                self._get_verification_date()

        # originating centre
        # TODO #574 Expand to include sub-centre
        self.extra_keys['_originatingCentre'] = CENTRE_TITLES.get(
            centre, "unknown centre %s" % centre)

        # forecast time unit as a cm string
        # TODO #575 Do we want PP or GRIB style forecast delta?
        self.extra_keys['_forecastTimeUnit'] = self._timeunit_string()

        # shape of the earth
        soe_code = self.shapeOfTheEarth
        # As this class is now *only* for GRIB1, 'shapeOfTheEarth' is not a
        # value read from the actual file :  It is really a GRIB2 param, and
        # the value is merely what eccodes (gribapi) gives as the default.
        # This was always = 6, until eccodes 0.19, when it changed to 0.
        # See https://jira.ecmwf.int/browse/ECC-811
        # The two represent different sized spherical earths.
        if soe_code not in (6, 0):
            raise ValueError('Unexpected shapeOfTheEarth value =', soe_code)

        soe_code = 6
        # *FOR NOW* maintain the old behaviour (radius=6371229) in all cases,
        # for backwards compatibility.
        # However, this does not match the 'radiusOfTheEarth' default from the
        # gribapi so is probably incorrect (see above issue ECC-811).
        # So we may change this in future.

        if soe_code == 0:
            # New supposedly-correct default value, matches 'radiusOfTheEarth'.
            geoid = coord_systems.GeogCS(semi_major_axis=6367470)
        elif soe_code == 6:
            # Old value, does *not* match the 'radiusOfTheEarth' parameter.
            geoid = coord_systems.GeogCS(6371229)

        gridType = gribapi.grib_get_string(self.grib_message, "gridType")

        if gridType in ["regular_ll", "regular_gg", "reduced_ll",
                        "reduced_gg"]:
            self.extra_keys['_x_coord_name'] = "longitude"
            self.extra_keys['_y_coord_name'] = "latitude"
            self.extra_keys['_coord_system'] = geoid
        elif gridType == 'rotated_ll':
            # TODO: Confirm the translation from angleOfRotation to
            # north_pole_lon (usually 0 for both)
            self.extra_keys['_x_coord_name'] = "grid_longitude"
            self.extra_keys['_y_coord_name'] = "grid_latitude"
            southPoleLon = longitudeOfSouthernPoleInDegrees
            southPoleLat = latitudeOfSouthernPoleInDegrees
            self.extra_keys['_coord_system'] = \
                coord_systems.RotatedGeogCS(
                    -southPoleLat,
                    math.fmod(southPoleLon + 180.0, 360.0),
                    self.angleOfRotation, geoid)
        elif gridType == 'polar_stereographic':
            self.extra_keys['_x_coord_name'] = "projection_x_coordinate"
            self.extra_keys['_y_coord_name'] = "projection_y_coordinate"

            if self.projectionCentreFlag == 0:
                pole_lat = 90
            elif self.projectionCentreFlag == 1:
                pole_lat = -90
            else:
                raise TranslationError("Unhandled projectionCentreFlag")

            # Note: I think the grib api defaults LaDInDegrees to 60 for grib1.
            self.extra_keys['_coord_system'] = \
                coord_systems.Stereographic(
                    pole_lat, self.orientationOfTheGridInDegrees, 0, 0,
                    self.LaDInDegrees, ellipsoid=geoid)

        elif gridType == 'lambert':
            self.extra_keys['_x_coord_name'] = "projection_x_coordinate"
            self.extra_keys['_y_coord_name'] = "projection_y_coordinate"

            flag_name = "projectionCenterFlag"

            if getattr(self, flag_name) == 0:
                pole_lat = 90
            elif getattr(self, flag_name) == 1:
                pole_lat = -90
            else:
                raise TranslationError("Unhandled projectionCentreFlag")

            LambertConformal = coord_systems.LambertConformal
            self.extra_keys['_coord_system'] = LambertConformal(
                self.LaDInDegrees, self.LoVInDegrees, 0, 0,
                secant_latitudes=(self.Latin1InDegrees, self.Latin2InDegrees),
                ellipsoid=geoid)
        else:
            raise TranslationError("unhandled grid type: {}".format(gridType))

        if gridType in ["regular_ll", "rotated_ll"]:
            self._regular_longitude_common()
            j_step = self.jDirectionIncrementInDegrees
            if not self.jScansPositively:
                j_step = -j_step
            self._y_points = (np.arange(self.Nj, dtype=np.float64) * j_step +
                              self.latitudeOfFirstGridPointInDegrees)

        elif gridType in ['regular_gg']:
            # longitude coordinate is straight-forward
            self._regular_longitude_common()
            # get the distinct latitudes, and make sure they are sorted
            # (south-to-north) and then put them in the right direction
            # depending on the scan direction
            latitude_points = gribapi.grib_get_double_array(
                    self.grib_message, 'distinctLatitudes').astype(np.float64)
            latitude_points.sort()
            if not self.jScansPositively:
                # we require latitudes north-to-south
                self._y_points = latitude_points[::-1]
            else:
                self._y_points = latitude_points

        elif gridType in ["polar_stereographic", "lambert"]:
            # convert the starting latlon into meters
            cartopy_crs = self.extra_keys['_coord_system'].as_cartopy_crs()
            x1, y1 = cartopy_crs.transform_point(
                self.longitudeOfFirstGridPointInDegrees,
                self.latitudeOfFirstGridPointInDegrees,
                ccrs.Geodetic())

            if not np.all(np.isfinite([x1, y1])):
                raise TranslationError("Could not determine the first latitude"
                                       " and/or longitude grid point.")

            self._x_points = x1 + self.DxInMetres * np.arange(self.Nx,
                                                              dtype=np.float64)
            self._y_points = y1 + self.DyInMetres * np.arange(self.Ny,
                                                              dtype=np.float64)

        elif gridType in ["reduced_ll", "reduced_gg"]:
            self._x_points = self.longitudes
            self._y_points = self.latitudes

        else:
            raise TranslationError("unhandled grid type")

    def _regular_longitude_common(self):
        """Define a regular longitude dimension."""
        i_step = self.iDirectionIncrementInDegrees
        if self.iScansNegatively:
            i_step = -i_step
        self._x_points = (np.arange(self.Ni, dtype=np.float64) * i_step +
                          self.longitudeOfFirstGridPointInDegrees)
        if "longitude" in self.extra_keys['_x_coord_name'] and self.Ni > 1:
            if _longitude_is_cyclic(self._x_points):
                self.extra_keys['_x_circular'] = True

    def _get_processing_done(self):
        """Determine the type of processing that was done on the data."""

        processingDone = 'unknown'
        timeRangeIndicator = self.timeRangeIndicator
        default = 'time _grib1_process_unknown_%i' % timeRangeIndicator
        processingDone = TIME_RANGE_INDICATORS.get(timeRangeIndicator, default)

        return processingDone

    def _get_verification_date(self):
        reference_date_time = self._referenceDateTime

        # calculate start time
        time_range_indicator = self.timeRangeIndicator
        P1 = self.P1
        P2 = self.P2
        if time_range_indicator == 0:
            # Forecast product valid at reference time + P1 P1>0),
            # or Uninitialized analysis product for reference time (P1=0).
            # Or Image product for reference time (P1=0)
            time_diff = P1
        elif time_range_indicator == 1:
            # Initialized analysis product for reference time (P1=0).
            time_diff = P1
        elif time_range_indicator == 2:
            # Product with a valid time ranging between reference time + P1
            # and reference time + P2
            time_diff = (P1 + P2) * 0.5
        elif time_range_indicator == 3:
            # Average(reference time + P1 to reference time + P2)
            time_diff = (P1 + P2) * 0.5
        elif time_range_indicator == 4:
            # Accumulation (reference time + P1 to reference time + P2)
            # product considered valid at reference time + P2
            time_diff = P2
        elif time_range_indicator == 5:
            # Difference(reference time + P2 minus reference time + P1)
            # product considered valid at reference time + P2
            time_diff = P2
        elif time_range_indicator == 10:
            # P1 occupies octets 19 and 20; product valid at
            # reference time + P1
            time_diff = P1 * 256 + P2
        elif time_range_indicator == 51:
            # Climatological Mean Value: multiple year averages of
            # quantities which are themselves means over some period of
            # time (P2) less than a year. The reference time (R) indicates
            # the date and time of the start of a period of time, given by
            # R to R + P2, over which a mean is formed; N indicates the number
            # of such period-means that are averaged together to form the
            # climatological value, assuming that the N period-mean fields
            # are separated by one year. The reference time indicates the
            # start of the N-year climatology. N is given in octets 22-23
            # of the PDS. If P1 = 0 then the data averaged in the basic
            # interval P2 are assumed to be continuous, i.e., all available
            # data are simply averaged together. If P1 = 1 (the units of
            # time - octet 18, code table 4 - are not relevant here) then
            # the data averaged together in the basic interval P2 are valid
            # only at the time (hour, minute) given in the reference time,
            # for all the days included in the P2 period. The units of P2
            # are given by the contents of octet 18 and Table 4.
            raise TranslationError("unhandled grib1 timeRangeIndicator "
                                   "= 51 (avg of avgs)")
        elif time_range_indicator == 113:
            # Average of N forecasts (or initialized analyses); each
            # product has forecast period of P1 (P1=0 for initialized
            # analyses); products have reference times at intervals of P2,
            # beginning at the given reference time.
            time_diff = P1
        elif time_range_indicator == 114:
            # Accumulation of N forecasts (or initialized analyses); each
            # product has forecast period of P1 (P1=0 for initialized
            # analyses); products have reference times at intervals of P2,
            # beginning at the given reference time.
            time_diff = P1
        elif time_range_indicator == 115:
            # Average of N forecasts, all with the same reference time;
            # the first has a forecast period of P1, the remaining
            # forecasts follow at intervals of P2.
            time_diff = P1
        elif time_range_indicator == 116:
            # Accumulation of N forecasts, all with the same reference
            # time; the first has a forecast period of P1, the remaining
            # follow at intervals of P2.
            time_diff = P1
        elif time_range_indicator == 117:
            # Average of N forecasts, the first has a period of P1, the
            # subsequent ones have forecast periods reduced from the
            # previous one by an interval of P2; the reference time for
            # the first is given in octets 13-17, the subsequent ones
            # have reference times increased from the previous one by
            # an interval of P2. Thus all the forecasts have the same
            # valid time, given by the initial reference time + P1.
            time_diff = P1
        elif time_range_indicator == 118:
            # Temporal variance, or covariance, of N initialized analyses;
            # each product has forecast period P1=0; products have
            # reference times at intervals of P2, beginning at the given
            # reference time.
            time_diff = P1
        elif time_range_indicator == 123:
            # Average of N uninitialized analyses, starting at the
            # reference time, at intervals of P2.
            time_diff = P1
        elif time_range_indicator == 124:
            # Accumulation of N uninitialized analyses, starting at
            # the reference time, at intervals of P2.
            time_diff = P1
        else:
            raise TranslationError("unhandled grib1 timeRangeIndicator "
                                   "= %i" % time_range_indicator)

        # Get the timeunit interval.
        interval_secs = self._timeunit_seconds()
        # Multiply by start-offset and convert to a timedelta.
        #     NOTE: a 'float' conversion is required here, as time_diff may be
        #     a numpy scalar, which timedelta will not accept.
        interval_delta = datetime.timedelta(
            seconds=float(time_diff * interval_secs))
        # Return validity_time = (reference_time + start_offset*time_unit).
        return reference_date_time + interval_delta

    @property
    def bmdi(self):
        # Not sure of any cases where GRIB provides a fill value.
        # Default for fill value is None.
        return None

    def core_data(self):
        try:
            data = self._data
        except AttributeError:
            data = self.data
        return data

    def phenomenon_points(self, time_unit):
        """
        Return the phenomenon time point offset from the epoch time reference
        measured in the appropriate time units.

        """
        time_reference = '%s since epoch' % time_unit
        return cf_units.date2num(self._phenomenonDateTime, time_reference,
                                 cf_units.CALENDAR_GREGORIAN)

    def phenomenon_bounds(self, time_unit):
        """
        Return the phenomenon time bound offsets from the epoch time reference
        measured in the appropriate time units.

        """
        # TODO #576 Investigate when it's valid to get phenomenon_bounds
        time_reference = '%s since epoch' % time_unit
        unit = cf_units.Unit(time_reference, cf_units.CALENDAR_GREGORIAN)
        return [unit.date2num(self._periodStartDateTime),
                unit.date2num(self._periodEndDateTime)]


def _longitude_is_cyclic(points):
    """Work out if a set of longitude points is cyclic."""
    # Is the gap from end to start smaller, or about equal to the max step?
    gap = 360.0 - abs(points[-1] - points[0])
    max_step = abs(np.diff(points)).max()
    cyclic = False
    if gap <= max_step:
        cyclic = True
    else:
        delta = 0.001
        if abs(1.0 - gap / max_step) < delta:
            cyclic = True
    return cyclic


def _message_values(grib_message, shape):
    gribapi.grib_set_double(grib_message, 'missingValue', np.nan)
    data = gribapi.grib_get_double_array(grib_message, 'values')
    data = data.reshape(shape)

    # Handle missing values in a sensible way.
    mask = np.isnan(data)
    if mask.any():
        data = ma.array(data, mask=mask, fill_value=np.nan)
    return data


def _load_generate(filename):
    messages = GribMessage.messages_from_filename(filename)
    for message in messages:
        editionNumber = message.sections[0]['editionNumber']
        if editionNumber == 1:
            message_id = message._raw_message._message_id
            grib_fh = message._file_ref.open_file
            message = GribWrapper(message_id, grib_fh=grib_fh)
        elif editionNumber != 2:
            emsg = 'GRIB edition {} is not supported by {!r}.'
            raise TranslationError(emsg.format(editionNumber,
                                               type(message).__name__))
        yield message


def load_cubes(filenames, callback=None):
    """
    Returns a generator of cubes from the given list of filenames.

    Args:

    * filenames:
        One or more GRIB filenames to load from.

    Kwargs:

    * callback:
        Function which can be passed on to :func:`iris.io.run_callback`.

    Returns:
        A generator containing Iris cubes loaded from the GRIB files.

    """
    import iris.fileformats.rules as iris_rules
    grib_loader = iris_rules.Loader(_load_generate,
                                    {},
                                    load_convert)
    return iris_rules.load_cubes(filenames, callback, grib_loader)


def load_pairs_from_fields(grib_messages):
    """
    Convert an iterable of GRIB messages into an iterable of
    (Cube, Grib message) tuples.

    This capability can be used to filter out fields before they are passed to
    the load pipeline, and amend the cubes once they are created, using
    GRIB metadata conditions.  Where the filtering
    removes a significant number of fields, the speed up to load can be
    significant:

        >>> import iris
        >>> from iris_grib import load_pairs_from_fields
        >>> from iris_grib.message import GribMessage
        >>> filename = iris.sample_data_path('polar_stereo.grib2')
        >>> filtered_messages = []
        >>> for message in GribMessage.messages_from_filename(filename):
        ...     if message.sections[1]['productionStatusOfProcessedData'] == 0:
        ...         filtered_messages.append(message)
        >>> cubes_messages = load_pairs_from_fields(filtered_messages)
        >>> for cube, msg in cubes_messages:
        ...     prod_stat = msg.sections[1]['productionStatusOfProcessedData']
        ...     cube.attributes['productionStatusOfProcessedData'] = prod_stat
        >>> print(cube.attributes['productionStatusOfProcessedData'])
        0

    This capability can also be used to alter fields before they are passed to
    the load pipeline.  Fields with out of specification header elements can
    be cleaned up this way and cubes created:

        >>> from iris_grib import load_pairs_from_fields
        >>> cleaned_messages = GribMessage.messages_from_filename(filename)
        >>> for message in cleaned_messages:
        ...     if message.sections[1]['productionStatusOfProcessedData'] == 0:
        ...         message.sections[1]['productionStatusOfProcessedData'] = 4
        >>> cubes = load_pairs_from_fields(cleaned_messages)

    Args:

    * grib_messages:
        An iterable of :class:`iris_grib.message.GribMessage`.

    Returns:
        An iterable of tuples of (:class:`iris.cube.Cube`,
        :class:`iris_grib.message.GribMessage`).

    """
    import iris.fileformats.rules as iris_rules
    return iris_rules.load_pairs_from_fields(grib_messages, load_convert)


def save_grib2(cube, target, append=False):
    """
    Save a cube or iterable of cubes to a GRIB2 file.

    Args:

    * cube:
        The :class:`iris.cube.Cube`, :class:`iris.cube.CubeList` or list of
        cubes to save to a GRIB2 file.
    * target:
        A filename or open file handle specifying the GRIB2 file to save
        to.

    Kwargs:

    * append:
        Whether to start a new file afresh or add the cube(s) to the end of
        the file. Only applicable when target is a filename, not a file
        handle. Default is False.

    """
    messages = (message for _, message in save_pairs_from_cube(cube))
    save_messages(messages, target, append=append)


def save_pairs_from_cube(cube):
    """
    Convert one or more cubes to (2D cube, GRIB message) pairs.
    Returns an iterable of tuples each consisting of one 2D cube and
    one GRIB message ID, the result of the 2D cube being processed by the GRIB
    save rules.

    Args:

    * cube:
        A :class:`iris.cube.Cube`, :class:`iris.cube.CubeList` or
        list of cubes.

    """
    x_coords = cube.coords(axis='x', dim_coords=True)
    y_coords = cube.coords(axis='y', dim_coords=True)
    if len(x_coords) != 1 or len(y_coords) != 1:
        raise TranslationError("Did not find one (and only one) x or y coord")

    # Save each latlon slice2D in the cube
    for slice2D in cube.slices([y_coords[0], x_coords[0]]):
        grib_message = gribapi.grib_new_from_samples("GRIB2")
        _save_rules.run(slice2D, grib_message, cube)
        yield (slice2D, grib_message)


def save_messages(messages, target, append=False):
    """
    Save messages to a GRIB2 file.
    The messages will be released as part of the save.

    Args:

    * messages:
        An iterable of grib_api message IDs.
    * target:
        A filename or open file handle.

    Kwargs:

    * append:
        Whether to start a new file afresh or add the cube(s) to the end of
        the file. Only applicable when target is a filename, not a file
        handle. Default is False.

    """
    # grib file (this bit is common to the pp and grib savers...)
    if isinstance(target, str):
        grib_file = open(target, "ab" if append else "wb")
    elif hasattr(target, "write"):
        if hasattr(target, "mode") and "b" not in target.mode:
            raise ValueError("Target not binary")
        grib_file = target
    else:
        raise ValueError("Can only save grib to filename or writable")

    try:
        for message in messages:
            gribapi.grib_write(message, grib_file)
            gribapi.grib_release(message)
    finally:
        # (this bit is common to the pp and grib savers...)
        if isinstance(target, str):
            grib_file.close()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration tests for loading various production definitions.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests


from iris import load_cube


class TestPDT8(tests.IrisGribTest):
    def setUp(self):
        # Load from the test file.
        file_path = tests.get_data_path(
            ("GRIB", "time_processed", "time_bound.grib2")
        )
        self.cube = load_cube(file_path)

    def test_coords(self):
        # Check the result has main coordinates as expected.
        for name, shape, is_bounded in [
            ("forecast_reference_time", (1,), False),
            ("time", (1,), True),
            ("forecast_period", (1,), True),
            ("pressure", (1,), False),
            ("latitude", (73,), False),
            ("longitude", (96,), False),
        ]:
            coords = self.cube.coords(name)
            self.assertEqual(
                len(coords),
                1,
                "expected one {!r} coord, found {}".format(name, len(coords)),
            )
            (coord,) = coords
            self.assertEqual(
                coord.shape,
                shape,
                "coord {!r} shape is {} instead of {!r}.".format(
                    name, coord.shape, shape
                ),
            )
            self.assertEqual(
                coord.has_bounds(),
                is_bounded,
                "coord {!r} has_bounds={}, expected {}.".format(
                    name, coord.has_bounds(), is_bounded
                ),
            )

    def test_cell_method(self):
        # Check the result has the expected cell method.
        cell_methods = self.cube.cell_methods
        self.assertEqual(
            len(cell_methods),
            1,
            "result has {} cell methods, expected one.".format(
                len(cell_methods)
            ),
        )
        (cell_method,) = cell_methods
        self.assertEqual(cell_method.coord_names, ("time",))


if __name__ == "__main__":
    tests.main()
